# Notes

All source code samples were taken from the Horovod examples repository 
[here](https://github.com/uber/horovod/tree/master/examples) 
(last checked: February 19, 2019). The samples that work with MNIST data have been 
slightly modified. Our changes are limited to,

*  The data loading mechanism
*  A bit of code cleanup
*  A few additional comments pertaining to our custom data loading mechanism

**Note:** All newly added statements follow a comment beginning with `[HPCNS]`. All 
statements that demonstrate the use of Horovod follow a comment beginning with 
`[Horovod]` (as added by Horovod developers).

**Caution:** Where job submission scripts are available for both Python 2 and Python 3, please 
do not submit both Python 2 and Python 3 jobs simultaneously, as one of the jobs might fail. If 
you would like to try both, please run these in tandem such that the second job is started only
after the first is finished.

## Keras samples

The following Keras samples are included:

1.  `mnist.py`: A simple MNIST processing example with only the essential Horovod code 
for distributed training.
2.  `mnist_advanced.py`: This sample is primarily the same as `mnist.py`. However, a 
few more advanced Horovod features are used.

## PyTorch samples

The following PyTorch samples are included:

1.  `mnist.py`: Demonstrates distributed training using Horovod with PyTorch. A 
simple convolutional neural network is trained on the MNIST dataset.
2.  `synthetic_benchmark.py`: A benchmark that can be used to measure performance 
of PyTorch with Horovod without using any external dataset.

**Note:** The job scripts for JURECA are prefixed with `.` for these samples, so that 
these scripts do not appear in the directory listing. The reason for doing this is
that our testing revealed issues with multi-node training. As soon as the issue has 
been resolved, we'll make the scripts available.

## Tensorflow samples

The following Tensorflow samples are included:

1.  `mnist.py`: Demonstrates distributed training using Horovod with the low-level
Tensorflow API. A simple convolutional neural network is trained on the MNIST dataset.
2.  `mnist_estimator.py`: Demonstrates distributed training using Horovod with the
high-level Estimator API in Tensorflow. A simple convolutional neural network is 
trained on the MNIST dataset.
3.  `synthetic_benchmark.py`: A simple benchmark that can be used to measure performance 
of Tensorflow with Horovod without using any external dataset.
